use super::{archive, package, Handle};
use crate::db;

use std::{
    collections::HashMap,
    path::{Path, PathBuf},
    sync::{
        atomic::{AtomicU32, Ordering},
        Arc, Mutex,
    },
};

use sea_orm::{
    ActiveModelTrait, ColumnTrait, Condition, ConnectionTrait, DbConn, EntityTrait, JoinType,
    ModelTrait, NotSet, QueryFilter, QuerySelect, Related, RelationTrait, Set, TransactionTrait,
};
use tokio::{
    runtime,
    sync::{
        mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender},
        RwLock,
    },
};

pub enum RepoCommand {
    ParsePkg(i32, PathBuf),
}

pub struct RepoSharedState {
    pub repos_dir: PathBuf,
    pub conn: DbConn,
    pub rx: Mutex<UnboundedReceiver<RepoCommand>>,
    pub tx: UnboundedSender<RepoCommand>,
    pub repos: RwLock<HashMap<i32, (AtomicU32, Arc<Mutex<()>>)>>,
}

impl RepoSharedState {
    pub fn new(
        repos_dir: impl AsRef<Path>,
        conn: DbConn,
        repos: HashMap<i32, (AtomicU32, Arc<Mutex<()>>)>,
    ) -> Self {
        let (tx, rx) = unbounded_channel();

        Self {
            repos_dir: repos_dir.as_ref().to_path_buf(),
            conn,
            rx: Mutex::new(rx),
            tx,
            repos: RwLock::new(repos),
        }
    }
}

/// The actor is responsible for mutating the repositories. They receive their commands their
/// messages and process these commands in both a synchronous and asynchronous way.
pub struct RepoActor {
    rt: runtime::Handle,
    state: Arc<RepoSharedState>,
}

impl RepoActor {
    pub fn new(rt: runtime::Handle, state: Arc<RepoSharedState>) -> Self {
        Self {
            rt,
            state: Arc::clone(&state),
        }
    }

    /// Run the main actor loop
    pub fn run(self) {
        while let Some(msg) = {
            let mut rx = self.state.rx.lock().unwrap();
            rx.blocking_recv()
        } {
            match msg {
                RepoCommand::ParsePkg(repo, path) => {
                    let _ = self.parse_pkg(repo, path);

                    if self
                        .state
                        .repos
                        .blocking_read()
                        .get(&repo)
                        .map(|n| n.0.load(Ordering::SeqCst))
                        == Some(0)
                    {
                        // TODO sync
                    }
                }
            }
        }
    }

    /// Parse a queued package for the given repository.
    fn parse_pkg(&self, repo: i32, path: PathBuf) -> crate::Result<()> {
        let pkg = package::Package::open(&path)?;
        let pkg = self
            .rt
            .block_on(db::query::package::insert(&self.state.conn, repo, pkg))?;
        let dest_path = self
            .state
            .repos_dir
            .join(repo.to_string())
            .join(pkg.id.to_string());
        std::fs::rename(path, dest_path)?;

        tracing::info!(
            "Added '{}-{}-{}' to repository {}",
            pkg.name,
            pkg.version,
            pkg.arch,
            repo,
        );

        self.state.repos.blocking_read().get(&repo).inspect(|n| {
            n.0.fetch_sub(1, Ordering::SeqCst);
        });

        Ok(())
    }

    fn sync_repo(&self, repo: i32) -> crate::Result<()> {
        let repos = self.state.repos.blocking_read();

        if let Some(_guard) = repos.get(&repo).map(|n| n.1.lock()) {
            let archs: Vec<String> = self.rt.block_on(
                db::Package::find()
                    .filter(db::package::Column::RepoId.eq(repo))
                    .select_only()
                    .column(db::package::Column::Arch)
                    .distinct()
                    .into_tuple()
                    .all(&self.state.conn),
            )?;

            for arch in archs {
                self.generate_archives(repo, &arch)?;
            }
        }

        Ok(())
    }

    fn generate_archives(&self, repo: i32, arch: &str) -> crate::Result<()> {
        Ok(())
    }
}
